// commandline.cpp

#include "stdafx.h"
#include "avesta.hpp"
#include "object.hpp"

using namespace avesta;

//==============================================================================

namespace
{
	class CommandLine : public Root< implements<ICommandLine> >
	{
	private:
		struct Argv
		{
			PTSTR Option;
			PTSTR Value;
		};
		std::vector<TCHAR>	m_text;
		std::vector<Argv>	m_args;
		std::vector<Argv>::iterator m_iter;

	public:
		CommandLine(PCTSTR commandLine)
		{
			m_text.assign(commandLine, commandLine + str::length(commandLine)+1);
			if(!m_text.empty())
				ParseArgv(&m_text[0], m_args);
			Reset();
		}
		bool Next(PTSTR* option, PTSTR* value)
		{
			if(m_iter == m_args.end())
				return false;
			*option = m_iter->Option;
			*value  = m_iter->Value;
			++m_iter;
			return true;
		}
		void Reset()
		{
			m_iter = m_args.begin();
		}

	private:
		static void ParseArgv(PTSTR commandline, std::vector<Argv>& args)
		{
			bool quoted = false;
			PTSTR beg, end;
			beg = end = commandline;
			while(true)
			{
				switch(*end)
				{
				case _T('\0'):
					// R}hCI
					if(beg < end)
						args.push_back(ParseOption(beg));
					return;
				case _T('\"'):
					if(quoted)
					{	// NH[gꂽ̒o
						*end = _T('\0');
						args.push_back(ParseOption(beg));
						quoted = false;
						beg = ++end;
					}
					else
					{	// NH[gJn
						quoted = true;
						beg = ++end;
					}
					break;
				case _T(' '):
					if(quoted)
					{	// NH[g̃Xy[X͕ƂĈ
						end++;
					}
					else
					{	// NH[gĂȂXy[X͋؂ƂĈ
						*end = _T('\0');
						if(beg < end)
							args.push_back(ParseOption(beg));
						beg = ++end;
					}
					break;
				default:
					// ͕̑ƂĈ
					end = str::inc(end);
				}
			}
		}
		static Argv ParseOption(PTSTR argv)
		{
			struct Argv arg;
			if(argv[0] == _T('-') || argv[0] == _T('/'))
			{	// IvVƒl̕
				TCHAR* strOption = argv+1;
				TCHAR* strValue = str::find_some_of(argv, _T(":="));
				if(strValue)
				{	// tIvV
					*strValue = _T('\0');
					strValue++;
					arg.Option = strOption;
					arg.Value  = strValue;
				}
				else
				{	// ȂIvV
					strValue++;
					arg.Option = strOption;
					arg.Value  = null;
				}
			}
			else
			{	// IvVȂGɃt@CpX
				arg.Option = null;
				arg.Value  = argv;
			}
			return arg;
		}
	};
}

//==============================================================================

ref<ICommandLine> avesta::ParseCommandLine(PCWSTR args)
{
	if(str::empty(args))
		return null;
	return objnew<CommandLine>(args);
}
