// ShellNotify.h
#pragma once

const int SHCNRF_InterruptLevel		= 0x0001;
const int SHCNRF_ShellLevel			= 0x0002;
const int SHCNRF_RecursiveInterrupt	= 0x1000;	/* Must be combined with SHCNRF_InterruptLevel */
const int SHCNRF_NewDelivery		= 0x8000;	/* Messages use shared memory */
 
class SHNotifyBase
{
protected:
	ULONG m_nSHChangeNotifyID;

public:
	SHNotifyBase() : m_nSHChangeNotifyID(0)
	{
	}
	~SHNotifyBase()
	{
		SHEndMonitor();
	}
	// ------------------------------------------------------
	//SHCNE_ALLEVENTS
	//SHCNE_ASSOCCHANGED
	//SHCNE_ATTRIBUTES
	//SHCNE_FREESPACE
	//SHCNE_UPDATEIMAGE

	//SHCNE_CREATE       : A nonfolder item has been created
	//SHCNE_DELETE       : A nonfolder item has been deleted.
	//SHCNE_MKDIR        : A folder has been created
	//SHCNE_RENAMEFOLDER : The name of a folder has changed (prev, next)
	//SHCNE_RENAMEITEM   : The name of a nonfolder item has changed.
	//SHCNE_RMDIR        : A folder has been removed (item, null)
	//SHCNE_UPDATEDIR    : The contents of an existing folder have changed, but the folder still exists and has not been renamed. 
	//SHCNE_UPDATEITEM   : An existing nonfolder item has changed, but the item still exists and has not been renamed. 

	//SHCNE_DRIVEADD
	//SHCNE_DRIVEADDGUI
	//SHCNE_DRIVEREMOVED
	//SHCNE_MEDIAINSERTED
	//SHCNE_MEDIAREMOVED
	//SHCNE_NETSHARE
	//SHCNE_NETUNSHARE
	//SHCNE_SERVERDISCONNECT
	//SHCNE_DISKEVENTS
	//SHCNE_GLOBALEVENTS
	//SHCNE_INTERRUPT

	// ------------------------------------------------------
	// fSources
	// SHCNRF_InterruptLevel
	// SHCNRF_ShellLevel
	// SHCNRF_RecursiveInterrupt
	// SHCNRF_NewDelivery
	//	SHChangeNotification_Lock
	//	SHChangeNotification_Unlock
	HRESULT SHBeginMonitor(int numEntries, SHChangeNotifyEntry entries[], HWND hWnd, UINT uMsg, LONG shcneMask, int shcnrf = SHCNRF_ShellLevel | SHCNRF_NewDelivery)
	{
		SHEndMonitor();
		m_nSHChangeNotifyID = ::SHChangeNotifyRegister(hWnd, shcnrf, shcneMask, uMsg, numEntries, entries);
		return m_nSHChangeNotifyID != 0 ? S_OK : E_FAIL;
	}
	HRESULT SHEndMonitor()
	{
		if(m_nSHChangeNotifyID == 0)
			return S_FALSE;
		BOOL res = ::SHChangeNotifyDeregister(m_nSHChangeNotifyID);
		m_nSHChangeNotifyID = 0;
		return res ? S_OK : E_FAIL;
	}
#ifdef _DEBUG
	void SHDumpEvent(LONG lEvent)
	{
		// VF̕ύXʒm
		if (lEvent & SHCNE_RENAMEITEM)
		{
			// ACê̖ύXꂽ
			TRACE(_T("OnSHChangeNotify($1) : SHCNE_RENAMEITEM"), m_nSHChangeNotifyID);
		}
		if (lEvent & SHCNE_CREATE)
		{
			// ACe쐬ꂽ
			TRACE(_T("OnSHChangeNotify($1) : SHCNE_CREATE"), m_nSHChangeNotifyID);
		}
		if (lEvent & SHCNE_DELETE)
		{
			// ACe폜ꂽ
			TRACE(_T("OnSHChangeNotify($1) : SHCNE_DELETE"), m_nSHChangeNotifyID);
		}
		if (lEvent & SHCNE_MKDIR)
		{
			// fBNg쐬ꂽ
			TRACE(_T("OnSHChangeNotify($1) : SHCNE_MKDIR"), m_nSHChangeNotifyID);
		}
		if (lEvent & SHCNE_RMDIR)
		{
			// fBNg폜ꂽ
			TRACE(_T("OnSHChangeNotify($1) : SHCNE_RMDIR"), m_nSHChangeNotifyID);
		}
		if (lEvent & SHCNE_MEDIAINSERTED)
		{
			TRACE(_T("OnSHChangeNotify($1) : SHCNE_MEDIAINSERTED"), m_nSHChangeNotifyID);
		}
		if (lEvent & SHCNE_MEDIAREMOVED)
		{
			TRACE(_T("OnSHChangeNotify($1) : SHCNE_MEDIAREMOVED"), m_nSHChangeNotifyID);
		}
		if (lEvent & SHCNE_DRIVEREMOVED)
		{
			TRACE(_T("OnSHChangeNotify($1) : SHCNE_DRIVEREMOVED"), m_nSHChangeNotifyID);
		}
		if (lEvent & SHCNE_DRIVEADD)
		{
			TRACE(_T("OnSHChangeNotify($1) : SHCNE_DRIVEADD"), m_nSHChangeNotifyID);
		}
		if (lEvent & SHCNE_NETSHARE)
		{
			TRACE(_T("OnSHChangeNotify($1) : SHCNE_NETSHARE"), m_nSHChangeNotifyID);
		}
		if (lEvent & SHCNE_NETUNSHARE)
		{
			TRACE(_T("OnSHChangeNotify($1) : SHCNE_NETUNSHARE"), m_nSHChangeNotifyID);
		}
		if (lEvent & SHCNE_ATTRIBUTES)
		{
			TRACE(_T("OnSHChangeNotify($1) : SHCNE_ATTRIBUTES"), m_nSHChangeNotifyID);
		}
		if (lEvent & SHCNE_UPDATEDIR)
		{
			TRACE(_T("OnSHChangeNotify($1) : SHCNE_UPDATEDIR"), m_nSHChangeNotifyID);
		}
		if (lEvent & SHCNE_UPDATEITEM)
		{
			// tH_EACȇύXꂽ
			TRACE(_T("OnSHChangeNotify($1) : SHCNE_UPDATEITEM"), m_nSHChangeNotifyID);
		}
		if (lEvent & SHCNE_SERVERDISCONNECT)
		{
			TRACE(_T("OnSHChangeNotify($1) : SHCNE_SERVERDISCONNECT"), m_nSHChangeNotifyID);
		}
		if (lEvent & SHCNE_UPDATEIMAGE)
		{
			TRACE(_T("OnSHChangeNotify($1) : SHCNE_UPDATEIMAGE"), m_nSHChangeNotifyID);
		}
		if (lEvent & SHCNE_DRIVEADDGUI)
		{
			TRACE(_T("OnSHChangeNotify($1) : SHCNE_DRIVEADDGUI"), m_nSHChangeNotifyID);
		}
		if (lEvent & SHCNE_RENAMEFOLDER)
		{
			// tH_̖̂ύXꂽ
			TRACE(_T("OnSHChangeNotify($1) : SHCNE_RENAMEFOLDER"), m_nSHChangeNotifyID);
		}
		if (lEvent & SHCNE_FREESPACE)
		{
			// t[Xy[X
			TRACE(_T("OnSHChangeNotify($1) : SHCNE_FREESPACE"), m_nSHChangeNotifyID);
		}
	}
#endif
};

template < class T >
class __declspec(novtable) SHNotifySink : public SHNotifyBase
{
public:
	LRESULT OnSHChangeNotify(UINT /*uMsg*/, WPARAM wParam, LPARAM lParam, BOOL& /*bHandled*/)
	{
		LONG lEvent = 0;
		ITEMIDLIST** pidls = NULL;
		HANDLE hLock = ::SHChangeNotification_Lock((HANDLE)wParam, (DWORD)lParam, &pidls, &lEvent);
		//DEBUG_ONLY( SHDumpEvent(lEvent); );
		static_cast<T*>(this)->HandleShellChangeNotify(lEvent, pidls[0], pidls[1]);
		::SHChangeNotification_Unlock(hLock);
		return 0;
	}
	void HandleShellChangeNotify(LONG lEvent, const ITEMIDLIST* pidl1, const ITEMIDLIST* pidl2)
	{
	}
};
